#!/usr/bin/env python2
# -*- coding: utf-8 -*-
import os
import sys
import logging
import logging.handlers
import argparse
import ConfigParser
# from config import config, version, logger
import config as g

import wstund_client
import wstund_server


def load_default_config():
    # global config
    g.config = ConfigParser.RawConfigParser()
    g.config.add_section('wstund')
    g.config.add_section('client')
    g.config.add_section('server')
    g.config.set('wstund', 'role', 'server')
    g.config.set('wstund', 'debug', 'false')
    g.config.set('wstund', 'logfile', '/var/log/wstund.log')
    g.config.set('wstund', 'loglevel', 'INFO')
    g.config.set('wstund', 'tundev', '/dev/net/tun')
    g.config.set('wstund', 'pidpath', '/run/wstund.pid')
    g.config.set('wstund', 'pidtimeout', '5')
    g.config.set('client', 'host', '127.0.0.1')
    g.config.set('client', 'port', '5000')
    g.config.set('client', 'ip', '10.10.0.4')
    g.config.set('client', 'netmask', '255.255.255.0')
    g.config.set('client', 'mtu', '1450')
    g.config.set('client', 'reconnect.interval', '5')
    g.config.set('client', 'script.up', None)
    g.config.set('client', 'script.down', None)
    g.config.set('server', 'host', '0.0.0.0')
    g.config.set('server', 'port', '5000')
    g.config.set('server', 'ip', '10.10.0.1')
    g.config.set('server', 'netmask', '255.255.255.0')
    g.config.set('server', 'mtu', '1450')


def load_config(args=None):
    load_default_config()

    if args.config is not None:
        g.config.read(args.config)

    if args.debug is not None:
        g.config.set('wstund', 'debug', 'true' if args.debug else 'false')

    levels = {k: getattr(logging, k) for k in ['CRITICAL', 'FATAL', 'ERROR',
                                               'WARNING', 'INFO', 'DEBUG']}
    lvl = g.config.get('wstund', 'loglevel').upper()
    lvl = levels[lvl] if lvl in levels else None
    lvl = lvl if args.verbose is None else max([logging.DEBUG,
                                                logging.ERROR - (args.verbose * 10)])
    set_logging_level(lvl)


def dump_config(role_only=False):
    role = g.config.get('wstund', 'role')
    for s in g.config.sections():
        if s not in ['wstund', role]:
            continue
        print("[{}]".format(s))
        for o in g.config.options(s):
            print("  {}: {}".format(o, g.config.get(s, o)))


def set_logging_level(plevel=None):
    lvl = logging.ERROR if plevel is None else plevel     # default logging level
    g.logger = logging.getLogger('wstund')              # global logger
    g.logger.setLevel(lvl)
    formatter = logging.Formatter('%(asctime)s %(name)s:%(levelname)s %(message)s')
    logs = [logging.handlers.RotatingFileHandler(g.config.get('wstund', 'logfile'),
                                                 maxBytes=1024*1024, backupCount=10)]
    if g.config.get('wstund', 'debug').lower() == 'true':
        logs.append(logging.StreamHandler())
    map(lambda l: l.setFormatter(formatter), logs)
    map(lambda l: g.logger.addHandler(l), logs)


def main():
    parser = argparse.ArgumentParser(description='websocket tunnel daemon')
    parser.add_argument('-c', '--config', nargs='?', default='/etc/wstund.conf',
                        help='specify config file')
    parser.add_argument('-r', '--role', default=None, choices=['server', 'client'],
                        help='specify role setting')
    parser.add_argument('-H', '--host', default=None, nargs='?',
                        help='specify host to listen/connect')
    parser.add_argument('-p', '--port', default=None, nargs='?',
                        help='specify port to listen/connect')
    parser.add_argument('-d', '--debug', action='store_true', default=None,
                        help='debug print out to current tty')
    parser.add_argument('-s', '--show_config', action='store_true', default=None,
                        help='show current configuration')
    parser.add_argument('-v', '--verbose', action='count', default=None,
                        help='increase log level, more times more logs')
    parser.add_argument('-V', '--version', action='version',
                        version='{}'.format(g.version),
                        help='show version infomation')

    args = parser.parse_known_args()[0]
    load_config(args)

    role = args.role if args.role is not None else g.config.get('wstund', 'role')
    if role not in ['server', 'client']:
        print("Please specify which role for wstund [server|client]")
        return
    g.config.set('wstund', 'role', role)

    def _set_args_configs(k):
        v = getattr(args, k)
        if v is not None:
            g.config.set(role, k, v)
    map(_set_args_configs, ['host', 'port'])
    g.logger.info('main role [{}]'.format(g.config.get('wstund', 'role')))

    if args.show_config is not None and args.show_config:
        dump_config(role_only=True)
        return

    main_role = wstund_client.main if role == 'client' else wstund_server.main
    main_role()


def sudo_main():
    if os.getuid() != 0:
        os.execvp("sudo", ["sudo", "python2"] + sys.argv)
    main()


if __name__ == "__main__":
    sudo_main()

# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai



#!/usr/bin/python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai

import os
import sys
import wstund

if __name__ == "__main__":
    # print("uid {0}".format(os.getuid()))
    if os.getuid() is not 0:
        os.execvp("sudo", ["sudo", "python2"] + sys.argv)
    wstund.main()



# -*- coding: utf-8 -*-
import sys
import select
import config as g
from time import sleep
from subprocess import check_call
from lockfile import LockTimeout
from daemon.runner import DaemonRunner, DaemonRunnerStopFailureError
from threading import Thread
from pytun import TunTapDevice, IFF_TUN, IFF_NO_PI
from ws4py.client.threadedclient import WebSocketClient
from ws4py.exc import HandshakeError


# ## for daemon.runner
class wstundClientApp():
    def __init__(self):
        # ## for daemon.runner
        g.logger.debug('YMK in wstundClientApp __init__')
        inout = '/dev/tty' if g.config.get('wstund', 'debug').lower() == 'true' else '/dev/null'
        self.stdin_path = self.stdout_path = self.stderr_path = inout
        self.pidfile_path = g.config.get('wstund', 'pidpath')
        self.pidfile_timeout = int(g.config.get('wstund', 'pidtimeout'))
        self.reconnect_interval = int(g.config.get('client', 'reconnect.interval'))
        # ## wstundClient
        self.ws_tun_client = wstundClient()

    def __del__(self):
        g.logger.debug('YMK in wstundClientApp __del__')
        self.ws_tun_client.stop()

    def run(self):
        while True:
            try:
                self.ws_tun_client.start()
            except HandshakeError as e:
                self.ws_tun_client.stop()
                g.logger.info('wstund client wait {0} seconds to reconnect'.format(self.reconnect_interval))
                sleep(self.reconnect_interval)


class webSocketTunClient(WebSocketClient):
    def __init__(self, url, tun, protocols=None, extensions=None, heartbeat_freq=None,
                 ssl_options=None, headers=None):
        WebSocketClient.__init__(self, url, protocols, extensions, heartbeat_freq,
                                 ssl_options, headers=headers)
        self.tun = tun

    def opened(self):
        g.logger.info("wstund connected")
        if g.config.get('client', 'script.up') is not None:
            # g.logger.debug(g.config.get('client', 'script.up'))
            check_call(g.config.get('client', 'script.up'), shell=True)

    def closed(self, code, reason):
        g.logger.info(("wstund closed", code, reason))
        if g.config.get('client', 'script.down') is not None:
            check_call(g.config.get('client', 'script.down'), shell=True)

    def received_message(self, m):
        if m is not None:
            g.logger.debug("incoming len(m) {0}".format(len(m)))
            try:
                self.tun.write(m.data)
            except RuntimeError as e:
                pass


class wstundClient():
    def __init__(self):
        # g.logger.debug('YMK in wstundClient __init__')
        self.tun = None
        self.ws = None
        self.thread = None
        self.running = False

    def outgoing(self):
        # while True:
        #     g.logger.debug('YMK in wstundClient outgoing')
        #     sleep(1)
        self.thread_closing = False
        self.count = 0
        poller = select.epoll()
        poller.register(self.tun, select.EPOLLIN)
        while True:
            # sleep(0.50)
            if self.thread_closing is True:
                break
            events = poller.poll(2)
            for fd, flag in events:
                if fd is self.tun.fileno():
                    buf = self.tun.read(self.tun.mtu)
                    # print("data len(buf) {1} [{0}]".format(hexlify(buf), len(buf)))
                    g.logger.debug("outgoing len(buf) {0}".format(len(buf)))
                    self.count += 1
                    self.ws.send(buf, True)

    def start(self):
        self.running = True
        # ## tuntap
        if self.tun is None:
            self.tun = TunTapDevice(flags=IFF_TUN | IFF_NO_PI)
        g.logger.debug(self.tun.name)
        self.tun.addr = g.config.get('client', 'ip')
        # self.tun.dstaddr = '10.10.0.1'
        self.tun.netmask = g.config.get('client', 'netmask')
        self.tun.mtu = int(g.config.get('client', 'mtu'))
        self.tun.up()

        # ## websocket
        # ws = PingClient('ws://{0}:{1}/ws'.format(args.host, args.port), protocols=['http-only', 'chat'])
        self.url = 'ws://{0}:{1}/ws'.format(g.config.get('client', 'host'), g.config.get('client', 'port'))
        self.ws = webSocketTunClient(self.url, self.tun, protocols=['http-only', 'chat'], heartbeat_freq=2.0)
        try:
            self.ws.daemon = True
            self.ws.connect()
        except:
            return

        # ## thread
        if self.thread is None:
            self.thread = Thread(target=self.outgoing)
            self.thread.daemon = True
            self.thread.start()

        # ## run forever
        self.run()

    def stop(self):
        if self.running is False:
            return
        self.thread_closing = True
        if self.thread is not None:
            self.thread.join()
            self.thread = None
        if self.ws is not None and not self.ws.terminated:
            self.ws.close()
            self.ws.closed(1000, "client receive signal.SIGTERM)")
            self.ws = None
        if self.tun is not None:
            self.tun.down()
        self.running = False

    def run(self):
        self.ws.run_forever()
        # while True:
        #     g.logger.debug('YMK in wstundClient run')
        #    sleep(1)


def main():
    daemon_runner = DaemonRunner(wstundClientApp())
    try:
        daemon_runner.daemon_context.files_preserve = [h.stream for h in g.logger.handlers]
        daemon_runner.do_action()
    except (DaemonRunnerStopFailureError, LockTimeout) as e:
        g.logger.error(e)
        sys.exit(e)

# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai



# -*- coding: utf-8 -*-
import config as g
from os.path import abspath, join, dirname
import sys
import random
import select
import logging
import cherrypy
from time import sleep
from lockfile import LockTimeout
from daemon.runner import DaemonRunner, DaemonRunnerStopFailureError
from threading import Thread
from pytun import TunTapDevice, IFF_TUN, IFF_NO_PI
# from ws4py import configure_logger
from ws4py.server.cherrypyserver import WebSocketPlugin, WebSocketTool
from ws4py.websocket import WebSocket
from ws4py.messaging import TextMessage


class TunWebSocketHandler(WebSocket):
    tun = None

    def __init__(self, *args, **kwargs):
        super(TunWebSocketHandler, self).__init__(*args, **kwargs)
        self.thread = None
        self.thread_counter = 0
        self.thread_closing = False

    def background_send(self):
        poller = select.epoll()
        poller.register(TunWebSocketHandler.tun, select.EPOLLIN)
        while True:
            events = poller.poll(2)
            for fd, flag in events:
                if fd is TunWebSocketHandler.tun.fileno():
                    buf = TunWebSocketHandler.tun.read(TunWebSocketHandler.tun.mtu)
                    g.logger.debug("websocket-broadcast data len(buf) {}".format(len(buf)))
                    cherrypy.engine.publish('websocket-broadcast', buf, True)
                    # self.send(buf, True)

            # ## timeout then check thread_closing
            if self.thread_closing is True:
                g.logger.debug("background_send thread_closing")
                break

    def opened(self):
        g.logger.debug("TunWebSocketHandler opened")

        if TunWebSocketHandler.tun is None:
            g.logger.debug("TunWebSocketHandler new tun")
            tun = TunTapDevice(flags=IFF_TUN | IFF_NO_PI)
            tun.addr = g.config.get('server', 'ip')
            # tun.dstaddr = '10.10.0.2'
            tun.netmask = g.config.get('server', 'netmask')
            tun.mtu = int(g.config.get('server', 'mtu'))
            tun.up()
            TunWebSocketHandler.tun = tun

        if self.thread is None:
            g.logger.debug("TunWebSocketHandler new thread")
            self.thread = Thread(target=self.background_send)
            self.thread.daemon = True
            self.thread.start()
        self.thread_counter += 1

    def received_message(self, m):
        if m is not None:
            if m.is_binary:
                g.logger.debug("received_message: bin len {}".format(len(m)))
                TunWebSocketHandler.tun.write(m.data)
            else:
                g.logger.debug("received_message: msg {}".format(m))

    def closed(self, code, reason="A client drop it's connection"):
        cherrypy.engine.publish('websocket-broadcast', TextMessage(reason))
        g.logger.debug("TunWebSocketHandler closing thread")
        self.thread_counter -= 1
        if self.thread_counter is 0:
            self.thread_closing = True
            self.thread.join()
            g.logger.debug("TunWebSocketHandler closed thread")
            self.thread = None
            self.thread_closing = False


class Root(object):
    def __init__(self, host, port, ssl=False):
        self.host = host
        self.port = port
        self.scheme = 'wss' if ssl else 'ws'

    @cherrypy.expose
    def index(self):
        return """<html>
    <head>
      <script type='application/javascript' src='https://ajax.googleapis.com/ajax/libs/jquery/1.8.3/jquery.min.js'></script>
      <script type='application/javascript' src='js/cbuffer.js'></script>
      <script type='application/javascript'>
        $(document).ready(function() {

          // websocket = '%(scheme)s://%(host)s:%(port)s/ws';
          websocket = '%(scheme)s://' + window.location.host + '/ws';
          console.log(window.location);
          if (window.WebSocket) {
            ws = new WebSocket(websocket);
          }
          else if (window.MozWebSocket) {
            ws = MozWebSocket(websocket);
          }
          else {
            console.log('WebSocket Not Supported');
            return;
          }

          var Packet = function (src, dst, size, time) {
            this.src = src;
            this.dst = dst;
            this.size = size;
            this.time = time;
          }
          var pcbuffs = CBuffer(10);

          window.onbeforeunload = function(e) {
            ws.close(1000, 'wstund monitor leave');

            if(!e) e = window.event;
            e.stopPropagation();
            e.preventDefault();
          };
          ws.onmessage = function (evt) {
             // console.log("ws.onmessage packet size " + evt.data.size);
             var reader = new FileReader();
             reader.onload = function(event) {
                arrayBufferNew = this.result;
                pdata  = new Uint8Array(this.result);
                // console.log("data length " + pdata.length);
                /* for (i = 0; i < pdata.length; i++) {
                    console.log("data[ " + i + "] " + pdata[i]);
                } */

                src_ip = pdata[12] + '.' + pdata[13] + '.' + pdata[14] + '.' + pdata[15];
                dst_ip = pdata[16] + '.' + pdata[17] + '.' + pdata[18] + '.' + pdata[19];
                // console.log("src " + src_ip + " dst " + dst_ip);
                pcbuffs.push(new Packet(src_ip, dst_ip, pdata.length, Date.now()));
                // console.log("pcbuffs last " + pcbuffs.last().time);

                $('#pcbs').empty();
                for (i = 0; i < pcbuffs.length; i++) {
                    var p = pcbuffs.get(i);
                    if (p == undefined)
                        break;
                    var d = new Date(p.time);
                    var appstr = "<tr><td>" + p.src + "</td><td>";
                    appstr += p.dst + "</td><td>" + p.size + "</td><td>";
                    appstr += d + "</td></tr>";
                    $('#pcbs').append(appstr);
                }
             }
             reader.readAsArrayBuffer(evt.data)
          };
          ws.onopen = function() {
             // ws.send("%(username)s entered the room");
             console.log("ws.onopen could send messages");
             $('#status').text("wstund status: CONNECTED");
          };
          ws.onclose = function(evt) {
             $('#status').text("wstund status: DISCONNECTED");
          };
        });
      </script>
    </head>
    <body>
    <h2>wstund monitor (%(scheme)s://%(host)s:%(port)s)</h2>
    <h3 id='status'>wstund status: </h3>
    <table>
    <thead><tr><th>src</th><th>dst</th><th>len</th><th>time</th></tr></thead>
    <tbody id='pcbs'>
    </tbody>
    </table>
    </body>
    </html>
    """ % {'username': "wstund", 'host': self.host, 'port': self.port, 'scheme': self.scheme}

    @cherrypy.expose
    def ws(self):
        g.logger.debug("Handler created: {}".format(repr(cherrypy.request.ws_handler)))


# ## for daemon.runner
class wstundServertApp():
    def __init__(self):
        # ## for daemon.runner
        inout = '/dev/tty' if g.config.get('wstund', 'debug').lower() == 'true' else '/dev/null'
        self.stdin_path = self.stdout_path = self.stderr_path = inout
        self.pidfile_path = g.config.get('wstund', 'pidpath')
        self.pidfile_timeout = int(g.config.get('wstund', 'pidtimeout'))
        self.host = g.config.get('server', 'host')
        self.port = int(g.config.get('server', 'port'))

    def run(self):
        g.logger.info('server running at {}:{}'.format(self.host, self.port))
        # ## fro cherrypy
        # configure_logger(level=logging.DEBUG)
        cherrypy.config.update({'server.socket_host': self.host,
                                'server.socket_port': self.port,
                                'tools.staticdir.root': abspath(join(dirname(__file__), 'static'))})
        WebSocketPlugin(cherrypy.engine).subscribe()
        cherrypy.tools.websocket = WebSocketTool()

        cherrypy.quickstart(Root(self.host, self.port), '', config={
            '/ws': {
                'tools.websocket.on': True,
                'tools.websocket.handler_cls': TunWebSocketHandler
            },
            '/js': {
                'tools.staticdir.on': True,
                'tools.staticdir.dir': 'js'
            }
        })


def main():
    daemon_runner = DaemonRunner(wstundServertApp())
    try:
        daemon_runner.daemon_context.files_preserve = [h.stream for h in g.logger.handlers]
        daemon_runner.do_action()
    except (DaemonRunnerStopFailureError, LockTimeout) as e:
        g.logger.error(e)
        sys.exit(e)

# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai



#!/usr/bin/env python2
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai

from distutils.core import setup

setup(
    name='wstund',
    version='0.1.2a',
    author='yumaokao',
    author_email='ymkq9h@gmail.com',
    description='A WebSocket Tunnel Daemon.',
    long_description=open('README.txt').read(),
    license='LICENSE.txt',
    url='http://www.github.com',

    packages=['wstund'],
    install_requires=[
        'ws4py',
        'python-pytun',
        'lockfile',
        'python-daemon',
        'cherrypy'
    ],
    entry_points={
        'console_scripts': ['wstund=wstund.wstund:sudo_main']
    }
)
